/*
  File name        : cycle.h
  Date of creation : 1/5/2009
  Version          : 1.0
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains functions for profiling

  History          :
  - initial version
*/

#ifndef __CYCLE_H__
#define __CYCLE_H__

#include <assert.h>
#include <stdint.h>

#include <snd/cycle-rdtsc.h>

__BEGIN_DECLS

/**
   Cycle counter management functions
*/
extern int enable_cycle_counters_;
int enable_cycle_counters (void);

/**
   Cpu clock
*/
extern double cpu_clock;

/**
   Cycle counter structure.
*/
typedef struct __attribute__ ((__packed__)) {
    char *file; /**< File name */
    char *function; /**< Function function name */
    int64_t nb_cycles; /**< Number of cycles consumed by a function */
    float variance; /**< Variance */
    int last; /**< last index */
    int64_t *last_nb_cycles; /**< last counters */
    int nb_calls; /**< Number of function calls */
    pid_t tid; /**< Thread identifier */
    pthread_t ptid; /**< Posid thread identifier */
} cycle_counter_t;

/**
   Set the local cycle counter.
*/
#define CYCLE_COUNTER_TIC(name)                               \
    static __thread cycle_counter_t *_counter_##name = NULL;  \
    int64_t _tic_##name = 0;                                  \
    int _tic_state_##name = 1;                                \
    if (enable_cycle_counters_ < 0) enable_cycle_counters (); \
    if (enable_cycle_counters_ > 0) {                         \
        if (_counter_##name == NULL)                          \
            _counter_##name = create_cycle_counter (__FILE__, \
                *#name ? ":"#name":"  : __FUNCTION__);        \
        _tic_##name = rdtsc ();                               \
    }                                                         \
    do {} while (0)

/**
   Check the local cycle counter.
*/
#define CYCLE_COUNTER_TOC(name)                                        \
    assert (_tic_state_##name == 1);                                   \
    _tic_state_##name = 0;                                             \
    if ((enable_cycle_counters_ > 0) && (_counter_##name != NULL)) {   \
        int64_t _delta_##name;                                         \
        _counter_##name->nb_cycles += _delta_##name =                  \
            _counter_##name->last_nb_cycles[_counter_##name->last++] = \
            rdtsc () - _tic_##name;                                    \
        _counter_##name->variance +=                                   \
            (float)_delta_##name * (float)_delta_##name;               \
        _counter_##name->last %= enable_cycle_counters_;               \
        _counter_##name->nb_calls++;                                   \
    }                                                                  \
    do {} while (0)

/**
   Locally desactivate a cycle counter.
*/
#define CYCLE_COUNTER_DOWN(name)      \
    assert (_tic_state_##name == 1);  \
    _tic_state_##name = 0;            \
    if (enable_cycle_counters_ > 0) { \
        _tic_##name -= rdtsc ();      \
    }                                 \
    do {} while (0)

/**
   Locally resactivate a cycle counter.
*/
#define CYCLE_COUNTER_UP(name)        \
    assert (_tic_state_##name == 0);  \
    _tic_state_##name = 1;            \
    if (enable_cycle_counters_ > 0) { \
        _tic_##name += rdtsc ();      \
    }                                 \
    do {} while (0)

/**
   Create and initialize to zero a wall-clock timer (ie a cycle counter)
   into the global variable cycle_counters.

   @param file file name
   @param function function name
   
   @return pointer of the freshly created cycle counter
*/
cycle_counter_t *create_cycle_counter (const char *file, const char *function);

/**
   Display cycle counters.

   Output are sorted by total number of consumed ycles.
*/
void display_cycle_counters (void);

/**
    Local cycle counter
*/
#define CYCLE_COUNTER(name, statement) \
    do {                               \
        CYCLE_COUNTER_TIC (name);      \
        statement;                     \
        CYCLE_COUNTER_TOC (name);      \
    } while (0)

__END_DECLS

#endif /* __CYCLE_H__ */

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
