/*
  File name        : cycle.c
  Date of creation : 10/6/2009
  Version          : 1.0
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains tests for cycle counters.

  History          :
  - initial version
*/

/* depend: */
/* cflags: */
/* linker: -lcycle -lm -lpthread */

#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <snd/cycle.h>

#define N 100000
#define Nt 100

int pseudo_random (unsigned int seed)
{

    CYCLE_COUNTER_TIC();

    int i;

    /* Pseudo random sequence initialisation */
    unsigned int x1 = 1;
    unsigned int x2 = seed;
    unsigned int x3 = 1 - seed;

    /* Pseudo random sequence initial scrambling */
    for (i = 0; i < 1000; i++) {
        unsigned int state;

        state = ((x1 >> 3) ^ x1) & 1;
        x1 = (state << 30) | (x1 >> 1);

        state = ((x2 >> 3) ^ (x2 >> 2) ^ (x2 >> 1) ^ x2) & 1;
        x2 = (state << 30) | (x2 >> 1);

        state = ((x3 >> 4) ^ (x2 >> 2) ^ x2) & 1;
        x3 = (state << 30) | (x2 >> 1);
    }

    int r = (x1 ^ x2 ^ x3);

    CYCLE_COUNTER_TOC();

    return r;
}

void *thread_pseudo_random (void *param)
{
    int i;
    for (i = 1234; i < 1243; i++)
        pseudo_random (i);
    return param;
}

int pseudo_random_noloop (unsigned int seed)
{

    CYCLE_COUNTER_TIC();

    int i;

    /* Pseudo random sequence initialisation */
    unsigned int x1 = 1;
    unsigned int x2 = seed;
    unsigned int x3 = 1 - seed;

    CYCLE_COUNTER_DOWN();

    /* Pseudo random sequence initial scrambling */
    for (i = 0; i < 1000; i++) {
        unsigned int state;

        state = ((x1 >> 3) ^ x1) & 1;
        x1 = (state << 30) | (x1 >> 1);

        state = ((x2 >> 3) ^ (x2 >> 2) ^ (x2 >> 1) ^ x2) & 1;
        x2 = (state << 30) | (x2 >> 1);

        state = ((x3 >> 4) ^ (x2 >> 2) ^ x2) & 1;
        x3 = (state << 30) | (x2 >> 1);
    }

    CYCLE_COUNTER_UP();

    int r = (x1 ^ x2 ^ x3);

    CYCLE_COUNTER_TOC();

    return r;
}

int randn ()
{

    CYCLE_COUNTER_TIC();

    int i = 0;
    float x, y;

    int r = 0;
    for (i = 0; i < Nt; i++) {

        CYCLE_COUNTER_TIC(intern_while);

        do
            /* generate a signal which is average random in [0,1] */
            x = (float) rand () / ((float)RAND_MAX + 1);

        while (x == 0);     /* x should not be zero */

        CYCLE_COUNTER_TOC(intern_while);

        /* generate another signal which is average random in [0,1] */
        y = (float) rand () / ((float)RAND_MAX + 1);

        /* generate noise which average is 0, and variance is sigma */
        float z = sqrtf (-2. * logf (x)) * cosf (2 * M_PI * y);

        r = z * (1 << 31);
    }
    
    CYCLE_COUNTER_TOC();

    return r;
}

int main (int argc, char *argv[])
{

    CYCLE_COUNTER_TIC();

    int i;

    /* process arguments */
    int do_thread = 0;
    if ((argc == 2) && (strncmp (argv[1], "-thread", 7) == 0)) {
        do_thread = 1;
        argc--;
    }
    if (argc > 1) {
        printf ("usage: %s\n", argv[0]);
        exit (1);
    }

    pthread_t thread;
    if (do_thread) {
        printf ("pseudo_random: thread\n");
        pthread_create (&thread, NULL, thread_pseudo_random, NULL);
    }

    double sum_pr = 0;
    printf ("pseudo_random:");
    fflush (stdout);
    for (i = 0; i < N; i++)
        sum_pr += pseudo_random (i);
    printf (" done\n");

    printf ("pseudo_random_noloop:");
    fflush (stdout);
    for (i = 0; i < N; i++)
        sum_pr += pseudo_random_noloop (i);
    printf (" done\n");

    double sum_rd = 0;
    printf ("random:");
    fflush (stdout);
    for (i = 0; i < N; i++)
        sum_rd += randn ();
    printf (" done\n");

    CYCLE_COUNTER_TOC();

    display_cycle_counters ();

    return 0;
}

/* test: {} */
/* test: CYCLE_COUNTERS_=0 {} */
/* test: CYCLE_COUNTERS_=1 {} */
/* test: CYCLE_COUNTERS_=y {} */
/* test: CYCLE_COUNTERS_=5 {} */
/* test: CYCLE_COUNTERS_=y {} -thread */

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
