/*
  File name        : sqrtf.c
  Date of creation : 31/7/2013
  Version          : 1.0
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains tests for sqrt function

  History          :
  - initial version
*/

/* depend: */
/* cflags: */
/* linker: -lcycle -lm */

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <snd/cycle.h>

float my_sqrtf (float val)
{
    float res = 0;

    /* check error */
    if ((isnan (val)) || (val < 0)) {
        errno = EDOM;
        return NAN;
    }
    if (!isfinite (val))
        return val;

    /* compute magnitude */
    float x = val;
    int n = 1;
    if (x > 10) {
        while (x > 1) {
            n++;
            x = x / 10;
        }
        n--;
    } else if (x < 1) {
        while (x < 1) {
            n--;
            x = x * 10;
        }
    }

    /* compute incremental step */
    n = n / 2 + ((n < 0) ? -1 : (n > 0) ? 1 : 0);
    float step = 1;
    if (n < 0) {
        while (n < 0) {
            step /= 10;
            n++;
        }
    } else {
        while (n > 0) {
            step *=10;
            n--;
        }
    }

    /* compute square root */
    float tmp = 0;
    while (tmp * tmp <= val)
        tmp += step;

    int j;
    for (j = 0; j < 10; j++)
        tmp = res = (val / tmp + tmp) / 2;

    return res;
}

#define N 10

void test_sqrtf (int k, float (*f) (float))
{
    int i;

    float c, d;
    switch (k) {
    case -11: c = 1.23456789e-12f; d= 1e-14f; break;
    case -8: c = 1.23456789e-9f; d= 1e-11f; break;
    case -5: c = 1.23456789e-6f; d= 1e-8f; break;
    case -2: c = 1.23456789e-3f; d= 1e-5f; break;
    case -1: c = 1.23456789e-2f; d= 1e-4f; break;
    case 0: c = 1.23456789e-1f; d= 1e-3f; break;
    case 1: c = 1.23456789e0f; d= 1e-2f; break;
    case 4: c = 1.23456789e3f; d= 1e1f; break;
    case 7: c = 1.23456789e6f; d= 1e4f; break;
    case 10: c = 1.23456789e9f; d= 1e7f; break;
    case 13: c = 1.23456789e12f; d= 1e10f; break;
    case 16: c = 1.23456789e15f; d= 1e13f; break;
    default: return;
    }

    float a = 0.0f;
    printf ("test 1e%d\n", k);
    CYCLE_COUNTER_TIC();
    for (i = 0; i < N; i++)
        a += f (c + i * d);
    CYCLE_COUNTER_TOC();
    printf ("a: %g\n", a);
}

int main (int argc, char *argv[])
{

    /* process arguments */
    if (argc > 1) {
        printf ("usage: %s\n", argv[0]);
        exit (1);
    }

    setenv ("CYCLE_COUNTERS_", "24", 1);

    test_sqrtf (-11, my_sqrtf);
    test_sqrtf (-11, sqrtf);
    test_sqrtf (-8, my_sqrtf);
    test_sqrtf (-8, sqrtf);
    test_sqrtf (-5, my_sqrtf);
    test_sqrtf (-5, sqrtf);
    test_sqrtf (-2, my_sqrtf);
    test_sqrtf (-2, sqrtf);
    test_sqrtf (-1, my_sqrtf);
    test_sqrtf (-1, sqrtf);
    test_sqrtf (0, my_sqrtf);
    test_sqrtf (0, sqrtf);
    test_sqrtf (1, my_sqrtf);
    test_sqrtf (1, sqrtf);
    test_sqrtf (4, my_sqrtf);
    test_sqrtf (4, sqrtf);
    test_sqrtf (7, my_sqrtf);
    test_sqrtf (7, sqrtf);
    test_sqrtf (10, my_sqrtf);
    test_sqrtf (10, sqrtf);
    test_sqrtf (13, my_sqrtf);
    test_sqrtf (13, sqrtf);
    test_sqrtf (16, my_sqrtf);
    test_sqrtf (16, sqrtf);

    display_cycle_counters ();

    return 0;
}

/* test: CYCLE_COUNTERS_=1 {} */

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
